from pairtree import *

# Get the store 'factory'
f = PairtreeStorageFactory()

# For a new store it is very important to set uri_base
# This restriction can be relaxed if necessary
store = f.get_store(store_dir="data", uri_base="info:local/", shorty_length=2, hashing_type='md5')

# Creates a directory in the cwd called 'data' and adds some pairtree boilerplate
# files

# Create object - this method throws an exception if the object already
# exists
# Use get_object if you do not care whether it existed on disc and just want a
# object to write and read from.
foobar = store.get_object('foobar')
# foobar = store.get_object('foobar')

# Add a byte sequence
md5hash = foobar.add_bytestream('foobar.txt', """Can just be an in memory string of bytes""")

print "%s hash of foobar.txt = %s" % (md5hash)

# Faking a file handle to a large file we'd rather streamed than copy into
# memory
from StringIO import StringIO
handle_to_large_file = StringIO()
handle_to_large_file.write('Foo bar foo bar foo bar foo bar foo bar foo bar')
handle_to_large_file.seek(0)
# End fakery...

# Store the large file in the root of the pairtree object and
# call it 'large_file.txt'
md5hash = foobar.add_bytestream('large_file.txt', handle_to_large_file)

print "%s hash of large file = %s" % (md5hash)

print foobar.list_parts()

# Create a file on disc
fn = open('/tmp/foo.txt', 'w')
fn.write('bah and foo')
fn.close()

# stream file from on disc location into the pairtree object
# but put it into a 'data/mine' directory
print "%s file hash = %s" % (foobar.add_file('/tmp/foo.txt', 'data/mine'))

# create a new object to test 'split-end' handling
foobartoo = store.create_object('foobartoo')
print "foomanchu.txt - %s: %s" % (foobartoo.add_bytestream('foomanchu.txt', 'Nothing to see here. Move along'))

# Results should be mostly unchanged from before
# with the only addition being the data directory
print foobar.list_parts()

# Results should just be foomanchu.txt
print foobartoo.list_parts()

# create a new object to test recursive path deletion
# in objects
foobarthree = store.create_object('foobarthree')

store.delete_object('foobarthree')




#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

This is a convenience object, used as a proxy for an object inside a pairtree store.
As such, it shouldn't be instanciated directly.
"""

import os, sys, shutil

import codecs

import string

from storage_exceptions import *

class PairtreeStorageObject(object):
    """
    The important methods:

        -  add_bytestream(filename, bytestream, path=None, buffer_size=None):
        -. get_bytestream(filename, streamable=False, path=None):
        -  del_file(filename, path=None):
        -  list_parts(path=None):

    First, setup up a simple store in 'data' and get an object called 'bar'
    (which will be equivalent to 'http://example.org/bar')
    
    >>> from pairtree import PairtreeStorageFactory
    >>> factory = PairtreeStorageFactory()
    >>> store = factory.get_store(store_dir='data', uri_base='http://example.org/')
    >>> bar = store.get_object('bar')

    Now add a simple string to a file called 'foo.txt'

    >>> bar.add_bytestream('foo.txt', 'can be any sequence of bytes')
    >>> bar.list_parts()
    ['foo.txt']
    >>> 

    Adding buffered content from a file:

    >>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
    ...   bar.add_bytestream('Firefox_wallpaper.png', stream)
    ... 
    >>> 

    Adding the same file to magic/path/inside/object - paths are automatically created on
    demand.

    >>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
    ...   bar.add_bytestream('Firefox_wallpaper.png', stream, path='magic/path/inside/object')
    ... 
    >>> 

    Removing the first copy of that file, which was added to the wrong place:

    >>> bar.del_file('Firefox_wallpaper.png')
    >>> bar.list_parts()
    ['magic', 'foo.txt']
    >>> bar.list_parts('magic/path')
    ['inside']
    >>> bar.list_parts('magic/path/inside/object')
    ['Firefox_wallpaper.png']
    >>> 

    There are also some convenience methods:

        -  L{add_bytestream_by_path}(self, filepath, bytestream, buffer_size=None):
        -  L{del_file_by_path}(self, filepath, bytestream):
        -  L{get_bytestream_by_path}(self, filepath, streamable=False):

    The I{by_path} suffix means that you can give it the whole path as one, and it will
    try to figure out what is intended, for example, consider the png we placed in a nested
    directory earlier:

    >>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
    ...   bar.add_bytestream('Firefox_wallpaper.png', stream, path='magic/path/inside/object')
    ... 

    This can be written as:

    >>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
    ...   bar.add_bytestream_by_path('magic/path/inside/object/Firefox_wallpaper.png', stream)
    ... 

    Getting files from an object
    ============================

    The flag I{streamable} is key here - if this is set to True, then you will be passed
    a file handle, which you must remember to close or use the construct:

    >>> with bar.get_bytestream('foo.txt', streamable=True) as text:
    ...   print text.read()
    ... 
    >>>

    This is very useful for large files you wish to scan through, but do not need to hold
    in memory all at the same time.

    By setting streamable to False, the entire file is read into memory and returned:

    >>> print bar.get_bytestream('foo.txt')
    can be any sequence of bytes
    """
    def __init__(self, id, fs_store_client):
        """
        @param id: Identifier for pairtree object
        @type id: identifier
        @param fs_store_client: A reference to an instance of L{PairtreeStorageClient}
        @type fs_store_client: L{PairtreeStorageClient}
        """
        self.fs = fs_store_client
        if id.startswith(self.fs.uri_base):
            self.id = id[len(self.fs.uri_base):]
        else:
            self.id = id
        self.uri = "%s%s" % (self.fs.uri_base, id)

    def add_bytestream(self, filename, bytestream, path=None, buffer_size=None):
        """
        Add a string or file to a given filename within this object. a C{path} may
        be supplied to store the file within a subdirectory of the object.
        
        @param path: (Optional) subdirectory path to store file in
        @type path: Directory path
        @param filename: Name of the file to write to
        @type filename: filename
        @param bytestream: Either a string or a file-like object to read from
        @type bytestream: L{str}|L{file}
        @param buffer_size: (Optional) Used for streaming filelike objects - defines the size of the buffer
        to read in each cycle.
        @type buffer_size: L{int}
        """
        if buffer_size:
            return self.fs.put_stream(self.id, path, filename, bytestream, buffer_size)
        return self.fs.put_stream(self.id, path, filename, bytestream)

    def add_bytestream_by_path(self, filepath, bytestream, buffer_size=None):
        """
        Add a string or file to a given filename within this object.
        
        The following adds the contents of footxt into a file 'foo.txt' in a
        subdirectory of the object 'data', which may or may not have existed prior
        to this call:
        
        >>> object.add_bytestream_by_path('data/foo.txt', footxt)
        
        @param filepath: (Optional) path to store the file in
        @type filepath: path to a file
        @param bytestream: Either a string or a file-like object to read from
        @type bytestream: L{str}|L{file}
        @param buffer_size: (Optional) Used for streaming filelike objects - defines the size of the buffer
        to read in each cycle.
        @type buffer_size: L{int}
        """
        path, filename = os.path.split(filepath)
        if buffer_size:
            return self.add_bytestream(filename, bytestream, path, buffer_size)
        return self.add_bytestream(filename, bytestream, path)

    def get_bytestream(self, filename, streamable=False, path=None, appendable=False):
        """
        Reads a file from a pairtree object - If streamable is set to True,
        this returns the filehandle for that file, which must be C{close()}'d
        once finished with. In python 2.6 and above, this can be done easily:
        
        >>> with object.get_bytestream('image001.tif', True, 'data/images') as stream:
                # Do something with the C{stream} handle
                pass
        
        stream is closed at the end of a C{with} block
        
        If appendable is set to True, then the file is opened "wb+" and can accept writes.
        Otherwise, the file is opened read-only.
        
        @param path: (Optional) subdirectory path to retrieve file from
        @type path: Directory path
        @param filename: Name of the file to read in
        @type filename: filename
        @param streamable: If True, returns a filelike handle to C{read()} from - 
        I{remember to C{close()} the file!} If False, reads in the file into a 
        bytestring and return that instead.
        @type streamable: True|False
        @returns: Either L{file} or L{str}
        """
        if appendable:
            return self.fs.get_appendable_stream(self.id, path=path, stream_name=filename)
        else:
            return self.fs.get_stream(self.id, path=path, stream_name=filename, streamable=streamable)

    def get_bytestream_by_path(self, filepath, streamable=False, appendable=False):
        """
        As L{get_bytestream}, but can ask for a file via a path:
        
        >>> print object.get_bytestream('data/foo/mytext.txt')
        ............
        
        @param filepath: (Optional) path of the file inside the object
        @type filepath: path to a file
        @param streamable: If True, returns a filelike handle to C{read()} from - 
        I{remember to C{close()} the file!} If False, reads in the file into a 
        bytestring and return that instead.
        @type streamable: True|False
        @returns: Either L{file} or L{str}
        """
        path, filename = os.path.split(filepath)
        return self.get_bytestream(filename, streamable, path, appendable)

    def add_file(self, from_file_location, path=None, new_filename=None, buffer_size=None):
        """
        Adds a file from a given location. Currently, the copy is due via python buffering
        the read from one file to the other. Might be easily replaceable with a C{shutil.copy}
        at a later date.
        
        If no new filename is set, it will use the original filename
        
        Aside from this, it works in the same fasion as L{add_bytestream}
        
        @param from_file_location: File path to read the file from
        @type from_file_location: Directory path
        @param path: (Optional) subdirectory within object to store file in
        @type path: Directory path
        @param new_filename: Name of the file to write to
        @type new_filename: filename
        @param buffer_size: (Optional) Used for streaming filelike objects - defines the size of the buffer
        to read in each cycle.
        @type buffer_size: L{int}
        """
        if os.path.exists(from_file_location):
            if not new_filename:
                _, new_filename = os.path.split(from_file_location)
            fh = open(from_file_location, 'rb')
            if buffer_size:
                return self.fs.put_stream(self.id, path, new_filename, bytestream=fh, buffer_size=buffer_size)
            return self.fs.put_stream(self.id, path, new_filename, bytestream=fh)
            fh.close()
        else:
            raise FileNotFoundException

    def del_file(self, filename, path=None):
        """
        Delete a file from the object. 
        
        If path is set, it will attempt to delete from that subpath.
        
        @param filename: Name of the file to delete
        @type filename: filename
        @param path: (Optional) subdirectory within object to delete file from
        @type path: Directory path
        """
        return self.fs.del_stream(self.id, filename, path)

    def del_file_by_path(self, filepath):
        """
        Delete a file from the object using the filepath as a subpath within the object.
        
        Eg::
        
            object_root --  foo.txt
                            foo2.txt
                            data    --  image1.jpg
                                        image2.jpg
        
        >>> object.del_file_by_path('data/image2.jpg')
        >>>
        
        @param filepath: subdirectory filepath within object to delete
        @type filepath: Directory path
        """
        path, filename = os.path.split(filepath)
        return self.del_file(filename, path)

    def del_path(self, subpath, recursive=False):
        """
        Delete a subpath from the object, and can do so recursively (optional)
        If the path is found to be not "empty" (ie has not parts in it) and
        recursive is not True, then it will raise a L{PathIsNotEmptyException}
        @param path: subdirectory path to delete
        @type path: Directory path
        @param recursive: Whether the delete is recursive (think rm -rf)
        @type recursive: bool
        """
        return self.fs.del_path(self.id, subpath, recursive)

    def list_parts(self, path=None):
        """
        List all the parts of object's root.
        
        If path is supplied, the parts in that subdirectory are returned.
        
        If the subpath doesn't exist, a L{ObjectNotFoundException} will be raised.
        
        >>> object.list_parts('data/images')
        [ 'image001.tif', 'image....    ]
        
        @param path: (Optional) List the parts contained in C{path}'s subdirectory
        @type path: Directory path
        @returns: L{list}
        """
        return self.fs.list_parts(self.id, path)

    def isfile(self, filepath):
        """
        Returns True or False depending on whether the path is a file or not.
        
        If the file doesn't exist, False is returned.
        
        @param path: Path to be tested
        @type path: Directory path
        @returns: L{bool}
        """
        return self.fs.isfile(self.id, filepath)
        
    def isdir(self, filepath):
        """
        Returns True or False depending on whether the path is a subdirectory or not.
        
        If the path doesn't exist, False is returned.
        
        @param path: Path to be tested
        @type path: Directory path
        @returns: L{bool}
        """
        return self.fs.isdir(self.id, filepath)

    def stat(self, filepath):
        """
        Returns the os.stat for a given file, or False if the file doesn't exist
        
        @param id: id of the object
        @type id: string
        @param filepath: Path to be tested
        @type filepath: Directory path
        @returns L{posix.stat_result} or False
        """
        if self.isfile(filepath):
            return self.fs.stat(self.id, filepath)
        else:
            return False

    def id_to_dirpath(self):
        """
        Get the path to the top of this object

            -  I{"foobar://ark.1" --> "fo/ob/ar/+=/ar/k,/1"}

        @returns: A directory path to the object's root directory
        """
#        return os.sep.join(self._id_to_dir_list(id))
        return self.fs._id_to_dirpath(self.id)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
FS Pairtree storage - Reverse lookup
====================================

Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

This is an implementation of a reverse lookup index, using the pairtree path spec to 
record the link between local id and the id's that it corresponds to.

eg to denote issn:1234-1234 as being linked to a global id of "uuid:1e4f..."

-->  create a file at ROOT_DIR/pairtree_rl/is/sn/+1/23/4-/12/34/uuid+1e4f...

Note that the id it links to is recorded as a filename encoded as per the pairtree spec.

Usage
=====

>>> from pairtree import PairtreeReverseLookup
>>> rl = PairtreeReverseLookup(storage_dir="ROOT")

>>> rl["issn:1234-1234"].append("uuid:1e4f...")

>>> rl["issn:1234-1234"]
["uuid:1e4f"]

>>> rl["issn:1234-1234"] = ["id:1", "uuid:32fad..."]
>>>

Notes
=====

This was created to avoid certain race conditions I had with a pickled dictionary for this index.
A sqllite or similar lookup would also be effective, but this one relies solely on pairtree.
"""

from __future__ import with_statement

import os

from pairtree_path import id_encode, id_decode, get_id_from_dirpath, get_path_from_dirpath, id_to_dirpath

PAIRTREE_RL = "pairtree_rl"

class PairtreeReverseLookup_list(object):
  def __init__(self, rl_dir, id):
    self._rl_dir = rl_dir
    self._id = id
    self._dirpath = id_to_dirpath(self._id, self._rl_dir)
  
  def _get_ids(self):
    if os.path.isdir(self._dirpath):
      ids = []
      for f in os.listdir(self._dirpath):
        ids.append(id_decode(f))
      return ids
    else:
      return []
  
  def _add_id(self, new_id):
    if not os.path.exists(self._dirpath):
      os.makedirs(self._dirpath)
    enc_id = id_encode(new_id)
    if not os.path.isfile(enc_id):
      with open(os.path.join(self._dirpath, enc_id), "w") as f:
        f.write(new_id)
  
  def _exists(self, id):
    if os.path.exists(self._dirpath):
      return id_encode(id) in os.listdir(self._dirpath)
    else:
      return False
  
  def append(self, *args):
    [self._add_id(x) for x in args if not self._exists(x)]
  
  def __len__(self):
    return len(os.listdir(self._dirpath))
  
  def __repr__(self):
    return "ID:'%s' -> ['%s']" % (self._id, "','".join(self._get_ids()))
  
  def __str__(self):
    return self.__repr__()
  
  def __iter__(self):
    for f in self._get_ids():
      yield id_decode(f)
    
class PairtreeReverseLookup(object):
  def __init__(self, storage_dir="data"):
    self._storage_dir = storage_dir
    self._rl_dir = os.path.join(storage_dir, PAIRTREE_RL)
    self._init_store()
  
  def _init_store(self):
    if not os.path.isdir(self._storage_dir):
      os.makedirs(self._storage_dir)
  
  def __getitem__(self, id):
    return PairtreeReverseLookup_list(self._rl_dir, id)
  
  def __setitem__(self, id, value):
    id_c = PairtreeReverseLookup_list(self._rl_dir, id)
    if isinstance(list, value):
      id_c.append(*value)
    else:
      id_c.append(value)

  def __delitem__(self, id):
    dirpath = id_to_dirpath(id, self._rl_dir)
    if os.path.isdir(dirpath):
      for f in os.listdir(dirpath):
        os.remove(os.path.join(dirpath, f))
      os.removedirs(dirpath)  # will throw OSError if the dir cannot be removed.
      self._init_store() # just in case



# -*- coding: utf-8 -*-
import random, unittest, re
from pairtree import PairtreeStorageClient

from pairtree import ppath


PAIRTREE_STORAGE_DIR = "/tmp/pairtree"

class TestPairtree(unittest.TestCase):
    def i2p2i(self, id, target, label):
        ppath = self.pairtree._id_to_dir_list(id)[1:]
        self.assertEqual(ppath, target)
        #self.assertEqual( reverse it)

    def roundtrip(self, id, label):
        ppath = self.pairtree.id_encode(id)
        new_id = self.pairtree.id_decode(ppath)
        self.assertEqual(id, new_id)
        self.ppath_roundtrip(id, label)
        #self.assertEqual( reverse it)

    def ppath_roundtrip(self, id, label):
        pp = ppath.get_id_from_dirpath(ppath.id_to_dirpath(id))
        self.assertEqual(pp, id)

    def setUp(self):
        self.pairtree = PairtreeStorageClient('http://example.org', PAIRTREE_STORAGE_DIR, 2)

    def test_empty(self):
        pass
        #try:
        #    ppath = PairPath("")
        #    self.assertFalse(True, 'Empty id should raise exception')
        #except BadPairPath:
        #    pass

    def testabc(self):
        self.i2p2i('abc', ['ab','c','obj'], 'basic 3-char case')
    
    def testabc_roundtrip(self):
        self.roundtrip('abc', 'basic 3-char case - roundtrip')

    def testabc(self):
        self.i2p2i('abcd', ['ab','cd', 'obj'], 'basic 4-char case')

    def testabc_roundtrip(self):
        self.roundtrip('abcd', 'basic 4-char case - roundtrip')
    
    def testabc(self):
        self.i2p2i('abcd', ['ab','cd', 'obj'], 'basic 4-char case')
        
    def testabc_roundtrip(self):
        self.roundtrip('abcd', 'basic 4-char case - roundtrip')

    def testxy(self):
        self.i2p2i('xy', ['xy', 'obj'], '2-char edge case')
        
    def testxy_roundtrip(self):
        self.roundtrip('xy', '2-char edge case - roundtrip')

    def testz(self):
        self.i2p2i('z', ['z', 'obj'], '1-char edge case')
        
    def testz_roundtrip(self):
        self.roundtrip('z', '1-char edge case - roundtrip')
        
    def test12_986xy4(self):
        self.i2p2i('12-986xy4', ['12', '-9', '86', 'xy', '4', 'obj'], 'hyphen')
        
    def test12_986xy4_roundtrip(self):
        self.roundtrip('12-986xy4', 'hyphen - roundtrip')

    def test_13030_45xqv_793842495(self):
        self.i2p2i('13030_45xqv_793842495',
                   ['13', '03', '0_', '45', 'xq', 'v_', '79', '38', '42', '49', 
                    '5', 'obj'],
                   'long id with undescores')

    def test_13030_45xqv_793842495_roundtrip(self):
        self.roundtrip('13030_45xqv_793842495',
                   'long id with undescores - roundtrip')

    def test_ark_13030_xt12t3(self):
        self.i2p2i('ark:/13030/xt12t3',
                   ['ar', 'k+', '=1', '30', '30', '=x', 't1', '2t', '3', 'obj'],
                   'colons and slashes')

    def test_ark_13030_xt12t3_roundtrip(self):
        self.roundtrip('ark:/13030/xt12t3',
                   'colons and slashes - roundtrip')

    def test_space(self):
        self.i2p2i('hello world', ['he', 'll', 'o^', '20', 'wo', 'rl', 'd', 'obj'], 'space')
        
    def test_space_roundtrip(self):
        self.roundtrip('hello world', 'space - roundtrip')
        
    def test_slash(self):
        self.i2p2i('/', ['=', 'obj'], '1-separator-char edge case')

    def test_slash_roundtrip(self):
        self.roundtrip('/', '1-separator-char edge case - roundtrip')

    def test_urn(self):
        self.i2p2i('http://n2t.info/urn:nbn:se:kb:repos-1',
                   ['ht', 'tp', '+=', '=n', '2t', ',i', 'nf', 'o=', 'ur', 'n+', 
                    'nb', 'n+', 'se', '+k', 'b+', 're', 'po', 's-', '1', 'obj'],
                   'a URL with colons, slashes, and periods')

    def test_urn_roundtrip(self):
        self.roundtrip('http://n2t.info/urn:nbn:se:kb:repos-1',
                   'a URL with colons, slashes, and periods - roundtrip')

    def test_wtf(self):
        self.i2p2i('what-the-*@?#!^!?',
                   ['wh', 'at', '-t', 'he', '-^', '2a', '@^', '3f', '#!', '^5', 
                    'e!', '^3', 'f', 'obj'],
                   'weird chars from spec example');

    def test_wtf_roundtrip(self):
        self.roundtrip('what-the-*@?#!^!?',
                   'weird chars from spec example - roundtrip');

    def test_weird(self):
        self.i2p2i('\\"*+,<=>?^|',
                   ['^5', 'c^', '22', '^2', 'a^', '2b', '^2', 'c^', '3c', '^3',
                    'd^', '3e', '^3', 'f^', '5e', '^7', 'c', 'obj'],
                   'all weird visible chars');

    def test_weird_roundtrip(self):
        self.roundtrip('\\"*+,<=>?^|',
                   'all weird visible chars - roundtrip');

    def test_basic_roundtrip(self):
        self.roundtrip('asdfghjklpoiuytrewqxcvbnm1234567890:;/', 'Basic Roundtrip')

    def test_french_roundtrip(self):
        self.roundtrip(u'Années de Pèlerinage', 'French Unicode roundtrip')

    def test_japanese_rountrip(self):
        self.roundtrip(u'ウインカリッスの日本語', 'Japanese Unicode roundtrip')
        
    def test_hardcore_unicode_rountrip(self):
        # If this works...
        self.roundtrip(u"""   1. Euro Symbol: €.
   2. Greek: Μπορώ να φάω σπασμένα γυαλιά χωρίς να πάθω τίποτα.
   3. Íslenska / Icelandic: Ég get etið gler án þess að meiða mig.
   4. Polish: Mogę jeść szkło, i mi nie szkodzi.
   5. Romanian: Pot să mănânc sticlă și ea nu mă rănește.
   6. Ukrainian: Я можу їсти шкло, й воно мені не пошкодить.
   7. Armenian: Կրնամ ապակի ուտել և ինծի անհանգիստ չըներ։
   8. Georgian: მინას ვჭამ და არა მტკივა.
   9. Hindi: मैं काँच खा सकता हूँ, मुझे उस से कोई पीडा नहीं होती.
  10. Hebrew(2): אני יכול לאכול זכוכית וזה לא מזיק לי.
  11. Yiddish(2): איך קען עסן גלאָז און עס טוט מיר נישט װײ.
  12. Arabic(2): أنا قادر على أكل الزجاج و هذا لا يؤلمني.
  13. Japanese: 私はガラスを食べられます。それは私を傷つけません。
  14. Thai: ฉันกินกระจกได้ แต่มันไม่ทำให้ฉันเจ็บ """,
                        "hardcore unicode test - roundtrip")

    def test_french(self):
        self.i2p2i('Années de Pèlerinage',
                   ['An', 'n^', 'c3', '^a', '9e', 's^', '20', 'de', '^2', '0P',
                    '^c', '3^', 'a8', 'le', 'ri', 'na', 'ge', 'obj'],
                   'UTF-8 chars')

        self.i2p2i("Années de Pèlerinage (Years of Pilgrimage) (S.160, S.161,\n\
 S.163) is a set of three suites by Franz Liszt for solo piano. Liszt's\n\
 complete musical style is evident in this masterwork, which ranges from\n\
 virtuosic fireworks to sincerely moving emotional statements. His musical\n\
 maturity can be seen evolving through his experience and travel. The\n\
 third volume is especially notable as an example of his later style: it\n\
 was composed well after the first two volumes and often displays less\n\
 showy virtuosity and more harmonic experimentation.",
                   ['An', 'n^', 'c3', '^a', '9e', 's^', '20', 'de', '^2', '0P',
                    '^c', '3^', 'a8', 'le', 'ri', 'na', 'ge', '^2', '0(', 'Ye',
                    'ar', 's^', '20', 'of', '^2', '0P', 'il', 'gr', 'im', 'ag', 
                    'e)', '^2', '0(', 'S,', '16', '0^', '2c', '^2', '0S', ',1',
                    '61', '^2', 'c^', '0a', '^2', '0S', ',1', '63', ')^', '20',
                    'is', '^2', '0a', '^2', '0s', 'et', '^2', '0o', 'f^', '20', 
                    'th', 're', 'e^', '20', 'su', 'it', 'es', '^2', '0b', 'y^', 
                    '20', 'Fr', 'an', 'z^', '20', 'Li', 'sz', 't^', '20', 'fo', 
                    'r^', '20', 'so', 'lo', '^2', '0p', 'ia', 'no', ',^', '20',
                    'Li', 'sz', 't\'', 's^', '0a', '^2', '0c', 'om', 'pl', 'et',
                    'e^', '20', 'mu', 'si', 'ca', 'l^', '20', 'st', 'yl', 'e^',
                    '20', 'is', '^2', '0e', 'vi', 'de', 'nt', '^2', '0i', 'n^',
                    '20', 'th', 'is', '^2', '0m', 'as', 'te', 'rw', 'or', 'k^',
                    '2c', '^2', '0w', 'hi', 'ch', '^2', '0r', 'an', 'ge', 's^',
                    '20', 'fr', 'om', '^0', 'a^', '20', 'vi', 'rt', 'uo', 'si',
                    'c^', '20', 'fi', 're', 'wo', 'rk', 's^', '20', 'to', '^2',
                    '0s', 'in', 'ce', 're', 'ly', '^2', '0m', 'ov', 'in', 'g^',
                    '20', 'em', 'ot', 'io', 'na', 'l^', '20', 'st', 'at', 'em',
                    'en', 'ts', ',^', '20', 'Hi', 's^', '20', 'mu', 'si', 'ca',
                    'l^', '0a', '^2', '0m', 'at', 'ur', 'it', 'y^', '20', 'ca',
                    'n^', '20', 'be', '^2', '0s', 'ee', 'n^', '20', 'ev', 'ol',
                    'vi', 'ng', '^2', '0t', 'hr', 'ou', 'gh', '^2', '0h', 'is',
                    '^2', '0e', 'xp', 'er', 'ie', 'nc', 'e^', '20', 'an', 'd^',
                    '20', 'tr', 'av', 'el', ',^', '20', 'Th', 'e^', '0a', '^2',
                    '0t', 'hi', 'rd', '^2', '0v', 'ol', 'um', 'e^', '20', 'is',
                    '^2', '0e', 'sp', 'ec', 'ia', 'll', 'y^', '20', 'no', 'ta',
                    'bl', 'e^', '20', 'as', '^2', '0a', 'n^', '20', 'ex', 'am',
                    'pl', 'e^', '20', 'of', '^2', '0h', 'is', '^2', '0l', 'at',
                    'er', '^2', '0s', 'ty', 'le', '+^', '20', 'it', '^0', 'a^',
                    '20', 'wa', 's^', '20', 'co', 'mp', 'os', 'ed', '^2', '0w',
                    'el', 'l^', '20', 'af', 'te', 'r^', '20', 'th', 'e^', '20',
                    'fi', 'rs', 't^', '20', 'tw', 'o^', '20', 'vo', 'lu', 'me',
                    's^', '20', 'an', 'd^', '20', 'of', 'te', 'n^', '20', 'di',
                    'sp', 'la', 'ys', '^2', '0l', 'es', 's^', '0a', '^2', '0s',
                    'ho', 'wy', '^2', '0v', 'ir', 'tu', 'os', 'it', 'y^', '20',
                    'an', 'd^', '20', 'mo', 're', '^2', '0h', 'ar', 'mo', 'ni',
                    'c^', '20', 'ex', 'pe', 'ri', 'me', 'nt', 'at', 'io', 'n,', 'obj'],                   
                   'very long id with apostrophes and UTF-8 chars')

    def test_id_to_url_simple(self):
        desired_url = "file://%s/pairtree_root/fo/o/obj/bar.txt" % PAIRTREE_STORAGE_DIR
        test_url = self.pairtree.get_url("foo", "bar.txt")
        self.assertEquals(desired_url, test_url)


    def test_id_to_url_withpath(self):
        desired_url = "file://%s/pairtree_root/fo/o/obj/data/subdir/bar.txt" % PAIRTREE_STORAGE_DIR
        test_url = self.pairtree.get_url("foo", "bar.txt", path="data/subdir")
        self.assertEquals(desired_url, test_url)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

This client handles all of the pairtree conventions, and provides a Pairtree object
to make it easier to interact with.

Usage
=====

>>> from pairtree import PairtreeStorageClient

To create a pairtree store in I{mystore/} to hold objects which have a URI base of
I{http://example.org/ark:/123}

>>> store = PairtreeStorageClient(store_dir='mystore', uri_base='http://example.org/ark:/123')

"""

import os, sys, shutil

import codecs

import string

import re

from storage_exceptions import *

from pairtree_object import PairtreeStorageObject

import pairtree_path as ppath

import hashlib

class PairtreeStorageClient(object):
    """A client that oversees the implementation of the Pairtree FS specification
    version 0.1.

    >>> from pairtree import PairtreeStorageClient
    >>> store = PairtreeStorageClient(store_dir='data', uri_base="http://")

    This will create the following on disc in a directory called 'data' if it doesn't already exist::

        $ ls -R data/
        data/:
        pairtree_prefix  pairtree_root  pairtree_version0_1

        data/pairtree_root:

    Where
        1. the file 'pairtree_prefix' contains just "http://"
        2. the file 'pairtree_version0_1' contains::

           This directory conforms to Pairtree Version 0.1.
           Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html

    Note, if data *had* already existed and was a pairtree store, the uri_base would
    have been read from the prefix file and override the one supplied above.

    Also, if you try to create a store over a directory that already exists, but which isn't
    a pairtree store that it can recognise, it will raise a L{NotAPairtreeStoreException}.
    """
    def __init__(self, uri_base, store_dir, shorty_length=2, hashing_type=None):
        """
        Constructor
        @param store_dir: The file directory where the pairtree store is
        @type store_dir: A path to a directory, relative or absolute
        @param uri_base: The URI base for the store
        @type uri_base: A URI fragment, like "http://example.org/"
        @param shorty_length: The size of the shorties in the pairtree implementation (Default: 2)
        @type shorty_length: integer
        @param hashing_type: The name of the algorithm to use when hashing files, if left as None, this is disabled.
        @type hashing_type: Any supported by C{hashlib}
        """
        self.store_dir = store_dir
        self.pairtree_root = os.path.join(self.store_dir, 'pairtree_root')
        self.uri_base = None
        if uri_base:
            self.uri_base = uri_base
        self.shorty_length = shorty_length
        self.hashing_type = hashing_type
        # regexes
        self._encode = re.compile(r"[\"*+,<=>?\\^|]|[^\x21-\x7e]", re.U)
        self._decode = re.compile(r"\^(..)", re.U)
        
        self._init_store()

    def __char2hex(self, m):
        return ppath.char2hex(m)

    def __hex2char(self, m):
        return ppath.hex2char(m)

    def _rm_uribase(self, id):
        if id.startswith(self.uri_base):
            return id[len(self.uri_base):]
        return id

    def id_encode(self, id):
        """
        The identifier string is cleaned of characters that are expected to occur rarely
        in object identifiers but that would cause certain known problems for file systems.
        In this step, every UTF-8 octet outside the range of visible ASCII (94 characters
        with hexadecimal codes 21-7e) [ASCII] (Cerf, “ASCII format for network interchange,”
        October 1969.), as well as the following visible ASCII characters::

           "   hex 22           <   hex 3c           ?   hex 3f
           *   hex 2a           =   hex 3d           ^   hex 5e
           +   hex 2b           >   hex 3e           |   hex 7c
           ,   hex 2c

        must be converted to their corresponding 3-character hexadecimal encoding, ^hh,
        where ^ is a circumflex and hh is two hex digits. For example, ' ' (space) is
        converted to ^20 and '*' to ^2a.

        In the second step, the following single-character to single-character conversions
        must be done::

               / -> =
               : -> +
               . -> ,

        These are characters that occur quite commonly in opaque identifiers but present
        special problems for filesystems. This step avoids requiring them to be hex encoded
        (hence expanded to three characters), which keeps the typical ppath reasonably
        short. Here are examples of identifier strings after cleaning and after
        ppath mapping::

            id:  ark:/13030/xt12t3
                ->  ark+=13030=xt12t3
                ->  ar/k+/=1/30/30/=x/t1/2t/3/
            id:  http://n2t.info/urn:nbn:se:kb:repos-1
                ->  http+==n2t,info=urn+nbn+se+kb+repos-1
                ->  ht/tp/+=/=n/2t/,i/nf/o=/ur/n+/n/bn/+s/e+/kb/+/re/p/os/-1/
            id:  what-the-*@?#!^!?
                ->  what-the-^2a@^3f#!^5e!^3f
                ->  wh/at/-t/he/-^/2a/@^/3f/#!/^5/e!/^3/f/

        (From section 3 of the Pairtree specification)

        @param id: Encode the given identifier according to the pairtree 0.1 specification
        @type id: identifier
        @returns: A string of the encoded identifier
        """
        return ppath.id_encode(id)

    def id_decode(self, id):
        """
        This decodes a given identifier from its pairtree filesystem encoding, into
        its original form:
        @param id: Identifier to decode
        @type id: identifier
        @returns: A string of the decoded identifier
        """
        return ppath.id_decode(id)

    def _get_id_from_dirpath(self, dirpath):
        """
        Internal - method for discovering the pairtree identifier for a
        given directory path.

        E.g.  pairtree_root/fo/ob/ar/+/  --> 'foobar:'

        @param dirpath: Directory path to decode
        @type dirpath: Path to object's root
        @returns: Decoded identifier
        """
        #path = self._get_path_from_dirpath(dirpath)
        #return self.id_decode("".join(path))
        return ppath.get_id_from_dirpath(dirpath, self.pairtree_root)

    def _get_path_from_dirpath(self, dirpath):
        """
        Internal - walks a directory chain and builds a list of the directory shorties
        relative to the pairtree_root

        @param dirpath: Directory path to walk
        @type dirpath: Directory path
        """
#        head, tail = os.path.split(dirpath)
#        path = [tail]
#        while not self.pairtree_root == head:
#            head, tail = os.path.split(head)
#            path.append(tail)
#        path.reverse()
#        return path
        return ppath.get_path_from_dirpath(dirpath, self.pairtree_root)


    def _id_to_dirpath(self, id):
        """
        Internal - method for turning an identifier into a pairtree directory tree
        of shorties.

            -  I{"foobar://ark.1" --> "fo/ob/ar/+=/ar/k,/1"}

        @param id: Identifer for a pairtree object
        @type id: identifier
        @returns: A directory path to the object's root directory
        """
        
#        return os.sep.join(self._id_to_dir_list(id))
        return ppath.id_to_dirpath(self._rm_uribase(id), self.pairtree_root, self.shorty_length)

    def _id_to_dir_list(self, id):
        """
        Internal - method for turning an identifier into a list of pairtree 
        directory tree of shorties.

            -  I{"foobar://ark.1" --> ["fo","ob","ar","+=","ar","k,","1"]}

        @param id: Identifer for a pairtree object
        @type id: identifier
        @returns: A list of directory path fragments to the object's root directory
        """
#        enc_id = self.id_encode(id)
#        dirpath = [self.pairtree_root]
#        while enc_id:
#            dirpath.append(enc_id[:self.shorty_length])
#            enc_id = enc_id[self.shorty_length:]
#        return dirpath
        return ppath.id_to_dir_list(self._rm_uribase(id), self.pairtree_root, self.shorty_length)
        
    def _init_store(self):
        """
        Initialise the store if the directory doesn't exist. Create the basic structure
        needed and write the prefix to disc.

        If the store directory exists, one of two things can happen:
            1. If that directory can be understood by this library as a pairtree store,
               it will attempt to read in the correct pairtree_prefix to use, instead of
               the supplied uri_base.
            2. If the directory cannot be understood, a L{NotAPairtreeStoreException} will
               be raised.
        """
        if not os.path.exists(self.store_dir):
            if self.uri_base:
                os.mkdir(self.store_dir)
                f = open(os.path.join(self.store_dir, "pairtree_version0_1"), "w")
                f.write("This directory conforms to Pairtree Version 0.1. Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html")
                f.close()
                f = open(os.path.join(self.store_dir, "pairtree_prefix"),"w")
                f.write(self.uri_base)
                f.close()
                os.mkdir(self.pairtree_root)
            else:
                raise NotAPairtreeStoreException("""No uri_base set for a non-existent
                                                    store - store cannot be instanciated""")
        else:
            if os.path.exists(os.path.join(self.store_dir, "pairtree_version0_1")):
                """Seems to be a pairtree0_1 compliant 'store'"""
                if os.path.exists(os.path.join(self.store_dir, "pairtree_prefix")):
                    """Read the uri base of this store"""
                    f = open(os.path.join(self.store_dir, "pairtree_prefix"),"r")
                    prefix = f.read().strip()
                    f.close()
                    self.uri_base = prefix
            else:
                raise NotAPairtreeStoreException

        if not os.path.isdir(self.store_dir):
            raise NotAPairtreeStoreException

    def list_ids(self):
        """
        Walk the store, and build a list of pairtree conformational objects in the
        store. This will return objects in 'split-ends' and will function correctly
        as long as non-shortie directorys are just that; non-shortie directories must
        have longer labels than the shorties - e.g::

              ab -- cd -- ef -- obj -- foo.txt
                     |     |
                     |     ---- gh
                     |           |
                     |           -- obj -- foo.txt
                     |
                     ---- e -- obj -- foo.txt

              This method will return ['abcdef', 'abcde', 'abcdefgh'] as ids in this
              store.
        
        Returns a generator, not a plain list since version 0.4.12

        @returns: L{generator}
        """

        objects = set()
        paths = [os.path.join(self.pairtree_root, x) for x in os.listdir(self.pairtree_root) if os.path.isdir(os.path.join(self.pairtree_root, x))]
        d = None
        terminator = ppath.get_terminator(self.shorty_length)
        if paths:
            d = paths.pop()
        while d:
            for t in os.listdir(d):
                if t == terminator:
                    potential_id = self._get_id_from_dirpath(os.path.join(d, terminator))
                    if potential_id not in objects:
                        objects.add(potential_id)
                        yield potential_id
                elif os.path.isdir(os.path.join(d, t)):
                    paths.append(os.path.join(d, t))
            if paths:
                d = paths.pop()
            else:
                d =False

    def _create(self, id):
        """
        Internal - create an object. If the object already exists, raise a
        L{ObjectAlreadyExistsException}

        @param id: Identifier to be created
        @type id: identifier
        @returns: L{PairtreeStorageObject}
        """
        id = self._rm_uribase(id)
        dirpath = os.path.join(self._id_to_dirpath(id))
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        else:
            raise ObjectAlreadyExistsException
        return PairtreeStorageObject(id, self)

    def list_parts(self, id, path=None):
        """
        List all the parts of the given identifer's parts (excluding shortie directories
        belonging to other objects)

        If path is supplied, the parts in that subdirectory are returned.

        If the subpath doesn't exist, a L{ObjectNotFoundException} will be raised.

        >>> store.list_parts('foobar:1', 'data/images')
        [ 'image001.tif', 'image....    ]

        @param id: Identifier for pairtree object
        @type id: identifier
        @param path: (Optional) List the parts contained in C{path}'s subdirectory
        @type path: Directory path
        @returns: L{list}
        """
        dirpath = os.path.join(self._id_to_dirpath(id))
        if path:
            dirpath = os.path.join(self._id_to_dirpath(id), path)
        if not os.path.exists(dirpath):
            raise ObjectNotFoundException
        return os.listdir(dirpath)
        #return [x for x in os.listdir(dirpath) if len(x)>self.shorty_length]

    def isfile(self, id, filepath):
        """
        Returns True or False depending on whether the path is a file or not.
        
        If the file doesn't exist, False is returned.
        
        @param filepath: Path to be tested
        @type filepath: Directory path
        @returns: L{bool}
        """
        dirpath = os.path.join(self._id_to_dirpath(id), filepath)
        try:
            return os.path.isfile(dirpath)
        except OSError:
            return False
        
    def isdir(self, id, filepath):
        """
        Returns True or False depending on whether the path is a subdirectory or not.
        
        If the path doesn't exist, False is returned.
        
        @param filepath: Path to be tested
        @type filepath: Directory path
        @returns: L{bool}
        """
        dirpath = os.path.join(self._id_to_dirpath(id), filepath)
        try:
            return os.path.isdir(dirpath)
        except OSError:
            return False
    
    def stat(self, id, filepath):
        """
        Returns the os.stat for a given file, or False if the file doesn't exist
        
        @param id: id of the object
        @type id: string
        @param filepath: Path to be tested
        @type filepath: Directory path
        @returns L{posix.stat_result} or False
        """
        if self.isfile(id, filepath):
            return os.stat(os.path.join(self._id_to_dirpath(id), filepath))
        else:
            return False

    def put_stream(self, id, path, stream_name, bytestream, buffer_size = 1024 * 64):
        """
        Store a stream of bytes into a file within a pairtree object.

        Can be either a string of bytes, or a filelike object which supports
        bytestream.read(buffer_size) - useful for very large files.

        @param id: Identifier for the pairtree object to write to
        @type id: identifier
        @param path: (Optional) subdirectory path to store file in
        @type path: Directory path
        @param stream_name: Name of the file to write to
        @type stream_name: filename
        @param bytestream: Either a string or a file-like object to read from
        @type bytestream: string|file
        @param buffer_size: (Optional) Used for streaming filelike objects - defines the size of the buffer
        to read in each cycle.
        @type buffer_size: integer
        @returns: tuple C{(hashing_algorithm, hash)} or None if hashing is disabled
        """
        dirpath = os.path.join(self._id_to_dirpath(id))
        if path:
            dirpath = os.path.join(self._id_to_dirpath(id), path)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        f = open(os.path.join(dirpath, stream_name), "wb")
        if self.hashing_type != None:
            hash_gen = getattr(hashlib, self.hashing_type)()
        try:
            # Stream file-like objects in with buffered reads
            if hasattr(bytestream, 'read'):
                try:
                    # try to get the stream back to zero
                    bytestream.seek(0)
                except:
                    pass
                if not buffer_size:
                    buffer_size = 1024 * 64
                chunk = bytestream.read(buffer_size)
                while chunk:
                    f.write(chunk)
                    if self.hashing_type != None:
                        hash_gen.update(chunk)
                    chunk = bytestream.read(buffer_size)
            else:
                f.write(bytestream)
                if self.hashing_type != None:
                    hash_gen.update(bytestream)
        except Exception, e:
            logger.info("put_stream failed: %s" % e)
        f.close()
        if self.hashing_type != None:
            return {"checksum":hash_gen.hexdigest(), "type":self.hashing_type}

    def get_appendable_stream(self, id, path, stream_name):
        """
        Reads a filehandle for a pairtree object. This is a "ab+" opened file and
        so can be appended to and obeys 'seek'
        
        >>> with store.get_appendable_stream('foobar:1','data/images', 'image001.tif') as stream:
                # Do something with the C{stream} handle
                pass

        stream is closed at the end of a C{with} block

        @param id: Identifier for the pairtree object to read from
        @type id: identifier
        @param path: (Optional) subdirectory path to retrieve file from
        @type path: Directory path
        @param stream_name: Name of the file to read in
        @type stream_name: filename
        @returns: L{file}
        """
        file_path = os.path.join(self._id_to_dirpath(id), stream_name)
        if path:
            file_path = os.path.join(self._id_to_dirpath(id), path, stream_name)
        f = open(file_path, "ab+")
        return f

    def get_url(self, id, stream_name, path=None):
        """
        Returns a direct 'file:///...' URL for a given id and stream (with an optional path)

        @param id: Identifier for the pairtree object to read from
        @type id: identifier
        @param path: (Optional) subdirectory path to the file location
        @type path: Directory path
        @param stream_name: Name of the file to point to
        @type stream_name: filename
        @returns: L{str} - eg "file:///opt/store/pairtree_root/..../foo.txt"
        """
        file_path = stream_name
        if path:
            file_path = os.path.join(path, stream_name)
        url_prefix = ppath.id_to_url(self._rm_uribase(id), self.pairtree_root, self.shorty_length)
        return os.sep.join((url_prefix, file_path))

    def get_stream(self, id, path, stream_name, streamable=False):
        """
        Reads a file from a pairtree object - If streamable is set to True,
        this returns the filehandle for that file, which must be C{close()}'d
        once finished with. In python 2.6 and above, this can be done easily:

        >>> with store.get_stream('foobar:1','data/images', 'image001.tif', True) as stream:
                # Do something with the C{stream} handle
                pass

        stream is closed at the end of a C{with} block

        @param id: Identifier for the pairtree object to read from
        @type id: identifier
        @param path: (Optional) subdirectory path to retrieve file from
        @type path: Directory path
        @param stream_name: Name of the file to read in
        @type stream_name: filename
        @param streamable: If True, returns a filelike handle to C{read()} from -
        I{remember to C{close()} the file!} If False, reads in the file into a
        bytestring and return that instead.
        @type streamable: True|False
        @returns: Either L{file} or L{str}
        """
        file_path = os.path.join(self._id_to_dirpath(id), stream_name)
        if path:
            file_path = os.path.join(self._id_to_dirpath(id), path, stream_name)
        if not os.path.exists(file_path):
            raise PartNotFoundException(id=id, path=path, stream_name=stream_name,file_path=file_path)
        f = open(file_path, "rb")
        if streamable:
            return f
        else:
            bytestream = f.read()
            f.close()
            return bytestream

    def del_stream(self, id, stream_name, path=None):
        """
        Delete a file from a pairtree object. Leaves no trace, be careful.
        @param id: Identifier for the pairtree object to delete from
        @type id: identifier
        @param path: (Optional) subdirectory path to delete file from
        @type path: Directory path
        @param stream_name: Name of the file to delete
        @type stream_name: filename
        """
        file_path = os.path.join(self._id_to_dirpath(id), stream_name)
        if path:
            file_path = os.path.join(self._id_to_dirpath(id), path, stream_name)
        if not os.path.exists(file_path):
            raise PartNotFoundException(id=id, path=path, stream_name=stream_name,file_path=file_path)
        if os.path.isdir(file_path):
            os.rmdir(file_path)
            isdir = True
        else:
            os.remove(file_path)
             
    def del_path(self, id, path, recursive=False):
        """
        Delete a subpath from an object, and can do so recursively (optional)
        If the path is found to be not "empty" (ie has not parts in it) and
        recursive is not True, then it will raise a L{PathIsNotEmptyException}
        @param id: Identifier for the pairtree object to delete from
        @type id: identifier
        @param path: subdirectory path to delete
        @type path: Directory path
        @param recursive: Whether the delete is recursive (think rm -r)
        @type recursive: bool
        """
        dirpath = os.path.join(self._id_to_dirpath(id), path)
        if not os.path.exists(dirpath):
            raise PartNotFoundException
        if os.path.isfile(dirpath):
            os.remove(dirpath)
        else:
            # It's a directory:
            all_parts = os.listdir(dirpath)
            
            if len(all_parts) == 0:
                os.rmdir(dirpath)
            elif recursive:
                # thankfully, terminators simplify this
                shutil.rmtree(dirpath)
            else:
                raise PathIsNotEmptyException

    def delete_object(self, id):
        """
        Delete's an object from the pairtree store, including any parts and subpaths
        There is no undo...
        @param id: Identifier of the object to delete
        @type id: identifier
        """
        dirs = self._id_to_dir_list(id)
        dirpath = os.path.join(os.sep.join(dirs))
        if not os.path.exists(dirpath):
            raise ObjectNotFoundException
        for item in self.list_parts(id):
            self.del_path(id,item, recursive=True)
        if not os.listdir(dirpath):
            os.rmdir(dirpath)
        # recursively delete up, if the directory is empty
        leaf = dirs.pop()
        while (not os.listdir(os.sep.join(dirs)) and os.sep.join(dirs) != self.pairtree_root):
            os.rmdir(os.sep.join(dirs))
            dirs.pop()

    def exists(self, id, path=None):
        """
        Answers the question "Does object or object subpath/file 'xxxxxxx' exist?"

        @param id: Identifier for the pairtree object to look for
        @type id: identifier
        @param path: Subpath or subfilepath to check
        @type path: Directory path
        @returns: L{bool}
        """
        dirpath = os.path.join(self._id_to_dirpath(id))
        if path:
            dirpath = os.path.join(self._id_to_dirpath(id), path)
        return os.path.exists(dirpath)

    def _get_new_id(self):
        """
        Inbuilt method to randomly generate an id, if one is not given to either
        L{get_object} or L{create_object}.

        Simply returns a random 14 digit long (base 10) number, not fantastically useful
        but at least makes sure it is unique in the store.

        @returns: Random but unique 14-digit long id number
        """
        id = "%.14d" % random.randint(0,99999999999999)
        while self.exists(id):
            id = "%.14d" % random.randint(0,99999999999999)
        return id

    def get_object(self, id=None, create_if_doesnt_exist=True):
        """
        Returns an pairtree object with identifier C{id} if it exists.

        If the object at C{id} doesn't exist then depending on C{create_if_doesnt_exist},

        >>> bar = client.get_object('bar')
        # the object with id 'bar' will be retrieved and created if necessary.

        Setting this flag to False, will cause it to raise an exception if it cannot find an object.

        >>> fake = client.get_object('doesnotexist', create_if_doesnt_exist=False)
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "build/bdist.linux-i686/egg/pairtree/pairtree_client.py", line 231, in get_object
        pairtree.storage_exceptions.ObjectNotFoundException

        (note that fake = client.get_object('doesnotexist', False) is equivalent to the above line)

        @param id: Identifier for the pairtree object to get (or create)
        @type id: identifier
        @param create_if_doesnt_exist: Flag - if True, an object will be created if it
        doesn't yet exist. Will raise an L{ObjectNotFoundException} if set to False
        and the object is non-existent.
        @type create_if_doesnt_exist: True|False
        @returns: L{PairtreeStorageObject}
        """
        if not id:
            id = self._get_new_id()
            return self._create(id)
        
        id = self._rm_uribase(id)
             
        if self.exists(id):
            return PairtreeStorageObject(id, self)
        elif create_if_doesnt_exist:
            return self._create(id)
        else:
            raise ObjectNotFoundException

    def create_object(self, id):
        """
        Creates a new object with identifier C{id}

        >>> bar = client.create_object('bar')
        >>>

        Note that reissuing that command again will raise an L{ObjectAlreadyExistsException}:

        >>> bar = client.create_object('bar')
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "build/bdist.linux-i686/egg/pairtree/pairtree_client.py", line 235, in create_object
        pairtree.storage_exceptions.ObjectAlreadyExistsException

        @param id: Identifier for the pairtree object to create
        @type id: identifier
        @returns: L{PairtreeStorageObject}
        """
        return self._create(id)




#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
FS Pairtree storage - Factory
=============================

Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

This is a simple Factory-style class, which can produce pairtree store clients.

Usage
=====

>>> from pairtree import PairtreeStorageFactory
>>> factory = PairtreeStorageFactory()

To create a pairtree store in I{mystore/} to hold objects which have a URI base of
I{http://example.org/ark:/123}

>>> store = factory.get_store(store_dir='mystore', uri_base='http://example.org/ark:/123')

"""

from pairtree_client import PairtreeStorageClient

class PairtreeStorageFactory(object):

    def get_store(self, store_dir="data", uri_base=None, shorty_length=2, hashing_type = None):
        """
        Get a store - if the store does not exist, one will be instanciated
        
        If hashing_type is set to one of the hashing algorithms supported by
        hashlib - ['md5', 'sha1', 'sha224','sha256','sha384','sha512'] then
        all bytestreams will be checksummed when added or updated and their sums returned.
        
        @param store_dir: The file directory where the pairtree store is
        @type store_dir: A path to a directory, relative or absolute
        @param uri_base: The URI base for the store
        @type uri_base: A URI fragment, like "http://example.org/"
        @param shorty_length: The size of the shorties in the pairtree implementation (Default: 2)
        @type shorty_length: integer
        @param hashing_type: The name of the algorithm to use when hashing files, if left as None, this is disabled.
        @type hashing_type: Any supported by C{hashlib}
        @returns: L{PairtreeStorageClient}
        """
        if hashing_type and hashing_type not in ['md5', 'sha1', 'sha224','sha256','sha384','sha512']:
            raise Exception("hashing type must be on of the supported hashlib types: md5, sha1, sha224, sha256, sha384, sha512")
        return PairtreeStorageClient(uri_base, store_dir, shorty_length, hashing_type)



#!/usr/bin/python
# -*- coding: utf-8 -*-


"""
FS Pairtree storage
===================

Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

NOTICE
======

The Pairtree specification on which this implementation is based is (c) 2009 UC Regents.

Various regexes used in path to id conversion and the bulk of the unittests were 
contributed by Erik Hetzner, based on John Kunze's work, also (c) 2009 UC Regents
and released under the Apache license.

The ppath script
================

A ppath script is included for convenience to be used in shell scripts or similar. Eg:

C{ppath topath} examples::

    $ vim mystore/pairtree_root/`ppath topath document:105/data/doc.txt`
    (Opens the file at mystore/pairtree_root/do/cu/me/nt/+1/05/data/doc.txt)
    $ cp `ppath topath foo:bar/1.txt` `ppath topath bar:foo/2.txt`

C{ppath toid} examples::

    data/subjects/pairtree_root/HA/SS/ET/ROOT$ ppath toid `pwd`
    HASSET/ROOT
    
Quick Start:
============

>>> from pairtree import *

>>> # Get the store 'factory'
>>> f = PairtreeStorageFactory()

The factory object is solely there to create clients for individual pairtree
stores. For example:

>>> store_foo = f.get_store(store_dir="data", uri_base="http://")

This will create the following on disc in a directory called 'data' if it doesn't already exist::

    $ ls -R data/
    data/:
    pairtree_prefix  pairtree_root  pairtree_version0_1

    data/pairtree_root:

Where 
    1. the file 'pairtree_prefix' contains just "http://"
    2. the file 'pairtree_version0_1' contains::
    
       This directory conforms to Pairtree Version 0.1.
       Updated spec: http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html

Note, if data *had* already existed and was a pairtree store, the uri_base would
have been read from the prefix file and override the one supplied above.

Also, if you try to create a store over a directory that already exists, but which isn't
a pairtree store that it can recognise, it will raise a NotAPairtreeStoreException.

Valid store names fit the regex ^[A-z][A-z0-9]* - but this is an arbitrary limitation
and can be removed if it is seen as unnecessary.

Creating and Getting store object:
==================================

Two main commands for this activity, eg continuing on:

>>> bar = store_foo.create_object('bar')
>>>

Note that reissuing that command again will raise an Exception:

>>> bar = store_foo.create_object('bar')
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "build/bdist.linux-i686/egg/pairtree/pairtree_client.py", line 235, in create_object
pairtree.storage_exceptions.ObjectAlreadyExistsException

There is also a 'get_object' command, which is more accommodating, as it can be passed
a fairly self-explanatory flag, which by default will create the object if it doesn't exist:

I{get_object(self, id=None, create_if_doesnt_exist=True)}

>>> bar = store_foo.get_object('bar')

Setting this flag to False, will cause it to raise an exception if it cannot find an object.

>>> fake = store_foo.get_object('doesnotexist', create_if_doesnt_exist=False)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "build/bdist.linux-i686/egg/pairtree/pairtree_client.py", line 231, in get_object
pairtree.storage_exceptions.ObjectNotFoundException

(note that fake = store_foo.get_object('doesnotexist', False) is equivalent to the above line)

A pairtree object:
==================

The important methods:

    -  add_bytestream(filename, bytestream, path=None, buffer_size=None):
    -. get_bytestream(filename, streamable=False, path=None):
    -  del_file(filename, path=None):
    -  list_parts(path=None):

E.g. - Examples speak louder than words

>>> bar.add_bytestream('foo.txt', 'can be any sequence of bytes')
>>> bar.list_parts()
['foo.txt']
>>> 

Adding buffered content from a file:

>>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
...   bar.add_bytestream('Firefox_wallpaper.png', stream)
... 
>>> 

Adding the same file to magic/path/inside/object - paths are automatically created on
demand.

>>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
...   bar.add_bytestream('Firefox_wallpaper.png', stream, path='magic/path/inside/object')
... 
>>> 

Removing the first copy of that file, which was added to the wrong place:

>>> bar.del_file('Firefox_wallpaper.png')
>>> bar.list_parts()
['magic', 'foo.txt']
>>> bar.list_parts('magic/path')
['inside']
>>> bar.list_parts('magic/path/inside/object')
['Firefox_wallpaper.png']
>>> 

There are also some convenience methods:

    -  add_bytestream_by_path(self, filepath, bytestream, buffer_size=None):
    -  del_file_by_path(self, filepath, bytestream):
    -  get_bytestream_by_path(self, filepath, streamable=False):

The I{by_path} suffix means that you can give it the whole path as one, and it will
try to figure out what is intended, for example, consider the png we placed in a nested
directory earlier:

>>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
...   bar.add_bytestream('Firefox_wallpaper.png', stream, path='magic/path/inside/object')
... 

This can be written as:

>>> with open('/home/ben/Firefox_wallpaper.png','rb') as stream:
...   bar.add_bytestream_by_path('magic/path/inside/object/Firefox_wallpaper.png', stream)
... 

Getting files from an object
============================

The flag I{streamable} is key here - if this is set to True, then you will be passed
a file handle, which you must remember to close or use the construct:

>>> with bar.get_bytestream('foo.txt', streamable=True) as text:
...   print text.read()
... 
>>>

This is very useful for large files you wish to scan through, but do not need to hold
in memory all at the same time.

By setting streamable to False, the entire file is read into memory and returned:

>>> print bar.get_bytestream('foo.txt')
can be any sequence of bytes
"""

__version__ = '0.5.2'

from pairtree_client import *
from pairtree_store import *
from pairtree_object import *
from pairtree_revlookup import PairtreeReverseLookup
import pairtree_path as ppath
from pairtree_path import id_encode, id_decode
from storage_exceptions import *

def id2path(id):
    """
    pass in a pairtree id and get back a path
    """
    path = ppath.id_to_dirpath(id)
    return path

def path2id(path):
    """
    pass in a pairtree path and get back an id
    """
    return ppath.get_id_from_dirpath(path)




class ObjectNotFoundException(Exception):
    """Object not found at the ID specified"""
    pass

class FileNotFoundException(Exception):
    """File cannot be found at the location requested"""
    pass

class PartNotFoundException(Exception):
    """Part not found"""
    def __init__(self, *p, **kw):
        self.context = (p, kw)
    def __str__(self):
        print " - Part not found: %s" % self.context

class StoreNotFoundException(Exception):
    """Store not found"""
    pass

class ObjectAlreadyExistsException(Exception):
    """Object ID already exists"""
    pass

class StoreAlreadyExistsException(Exception):
    """Store ID already exists"""
    pass

class PathIsNotEmptyException(Exception):
    """Cannot remove a path that isn't empty without the recursive flag set'"""
    pass

class NotAPairtreeStoreException(Exception):
    """The directory indicated exists, but doesn't
    announce itself to be a pairtree store via a
    'pairtree_version0_1' marker file in the root."""

class NotAValidStoreName(Exception):
    """Invalid name for a store. Must conform to ^[A-z][A-z0-9]* regex"""




#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Conventions used:

From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html version 0.1

This client handles all of the pairtree conventions, and provides a Pairtree object
to make it easier to interact with.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

EXTENSION - This version uses terminators to signal the directory within which an object resides. This
is an effort to avoid the 'split-end' issue and to simplify a number of issues to do with object file and subpath
listings.

So:

foobar:1/ goes to fo/ob/ar/+1/obj/

and 

bar:1/subpath/manifest.rdf goes to ba/r+/1/obj/subpath/manifest.rdf

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Usage
=====

>>> from pairtree import PairtreeStorageClient

To create a pairtree store in I{mystore/} to hold objects which have a URI base of
I{http://example.org/ark:/123}

>>> store = PairtreeStorageClient(store_dir='mystore', uri_base='http://example.org/ark:/123')

"""

import os, sys, shutil

import codecs

import string

import re

from storage_exceptions import *

import logging

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger('pairtreepath')

encode_regex = re.compile(r"[\"*+,<=>?\\^|]|[^\x21-\x7e]", re.U)
decode_regex = re.compile(r"\^(..)", re.U)

def char2hex(m):
        return "^%02x"%ord(m.group(0))

def hex2char(m):
        return chr(int(m.group(1), 16))


def id_encode(id):
    """
    The identifier string is cleaned of characters that are expected to occur rarely
    in object identifiers but that would cause certain known problems for file systems.
    In this step, every UTF-8 octet outside the range of visible ASCII (94 characters
    with hexadecimal codes 21-7e) [ASCII] (Cerf, “ASCII format for network interchange,”
    October 1969.), as well as the following visible ASCII characters::

        "   hex 22           <   hex 3c           ?   hex 3f
        *   hex 2a           =   hex 3d           ^   hex 5e
        +   hex 2b           >   hex 3e           |   hex 7c
        ,   hex 2c

    must be converted to their corresponding 3-character hexadecimal encoding, ^hh,
    where ^ is a circumflex and hh is two hex digits. For example, ' ' (space) is
    converted to ^20 and '*' to ^2a.

    In the second step, the following single-character to single-character conversions
    must be done::

            / -> =
            : -> +
            . -> ,

    These are characters that occur quite commonly in opaque identifiers but present
    special problems for filesystems. This step avoids requiring them to be hex encoded
    (hence expanded to three characters), which keeps the typical ppath reasonably
    short. Here are examples of identifier strings after cleaning and after
    ppath mapping::

        id:  ark:/13030/xt12t3
            ->  ark+=13030=xt12t3
            ->  ar/k+/=1/30/30/=x/t1/2t/3/obj/
        id:  http://n2t.info/urn:nbn:se:kb:repos-1
            ->  http+==n2t,info=urn+nbn+se+kb+repos-1
            ->  ht/tp/+=/=n/2t/,i/nf/o=/ur/n+/n/bn/+s/e+/kb/+/re/p/os/-1/obj/
        id:  what-the-*@?#!^!?
            ->  what-the-^2a@^3f#!^5e!^3f
            ->  wh/at/-t/he/-^/2a/@^/3f/#!/^5/e!/^3/f/obj/

    (From section 3 of the Pairtree specification - with obj extension)

    @param id: Encode the given identifier according to the pairtree 0.1 specification
    @type id: identifier
    @returns: A string of the encoded identifier
    """
    # Unicode or bust
    if isinstance(id, unicode):
        # assume utf-8
        # TODO - not assume encoding
        id = id.encode('utf-8')

    second_pass_m = {'/':'=',
                        ':':'+',
                        '.':','
                    }
    # hexify the odd characters
    # Using Erik Hetzner's regex in place of my previous hack
    new_id = encode_regex.sub(char2hex, id)
    
    # 2nd pass
    second_pass = []
    for char in new_id:
        second_pass.append(second_pass_m.get(char, char))
    return "".join(second_pass)

def id_decode(id):
    """
    This decodes a given identifier from its pairtree filesystem encoding, into
    its original form:
    @param id: Identifier to decode
    @type id: identifier
    @returns: A string of the decoded identifier
    """
    second_pass_m = {'=':'/',
                        '+':':',
                        ',':'.'
                    }
    second_pass = []
    for char in id:
        second_pass.append(second_pass_m.get(char, char))
    dec_id = "".join(second_pass)
    #dec_id = id.translate(string.maketrans(u'=+,',u'/:.'))
    # Using Erik Hetzner's regex in place of my previous hack
    #ppath_s = re.sub(r"\^(..)", self.__hex2char, dec_id)
    ppath_s = decode_regex.sub(hex2char, dec_id)
    # Again, drop the assumption of utf-8
    return ppath_s.decode('utf-8')

def get_terminator(shorty_length=2, end_of_pairtree_path="obj"):
    """
    Terminator directory name, based on the shorty length.
    
    It has to be one greater than the shorty length of a given pairtree, to avoid being
    part of a splitend and potentially incorporated or interpreted as part of another object
    
    shorty length = 1 (with default end of path string)
    -> "ob"
    shorty length = 2
    -> "obj"
    shorty length = 6
    -> "objobjo"
    """
    return "".join([(end_of_pairtree_path[x % len(end_of_pairtree_path)]) for x in xrange(shorty_length+1)])

def get_id_from_dirpath(dirpath, pairtree_root=""):
    """
    Internal - method for discovering the pairtree identifier for a
    given directory path.

    E.g.  pairtree_root/fo/ob/ar/+/  --> 'foobar:'

    @param dirpath: Directory path to decode
    @type dirpath: Path to object's root
    @returns: Decoded identifier
    """
    path, shorty_length = get_path_from_dirpath(dirpath, pairtree_root)
    assert path[-1] == get_terminator(shorty_length=shorty_length)
    return id_decode("".join(path[:-1]))

def get_path_from_dirpath(dirpath, pairtree_root=""):
    """
    Internal - walks a directory chain and builds a list of the directory shorties
    relative to the pairtree_root

    @param dirpath: Directory path to walk
    @type dirpath: Directory path
    """
    shorty_length = None
    head, tail = os.path.split(dirpath)
    path = [tail]
    while not pairtree_root == head:
        head, tail = os.path.split(head)
        shorty_length = len(tail)
        path.append(tail)
    path.reverse()
    try:
        assert path[-1] == get_terminator(shorty_length=shorty_length)
    except AssertionError:
        if len(path) == 2:
            # edge case - id length of 1. Using the terminator length as (shorty_length+1)
            shorty_length = len(path[-1])-1
        else:
            logger.error("For dirpath: %s and pairtree_root=%s" % (dirpath, pairtree_root))
            logger.error("Anticipated to get an %s at the end of the path, got %s instead" % (get_terminator(shorty_length=shorty_length), path) )
    return path, shorty_length

def id_to_dirpath(id, pairtree_root="", shorty_length=2):
    """
    Internal - method for turning an identifier into a pairtree directory tree
    of shorties.

        -  I{"foobar://ark.1" --> "fo/ob/ar/+=/ar/k,/1"}

    @param id: Identifer for a pairtree object
    @type id: identifier
    @returns: A directory path to the object's root directory
    """
    return os.sep.join(id_to_dir_list(id, pairtree_root, shorty_length))

def id_to_url(id, pairtree_root, shorty_length=2):
    """Method to provide 'file:///..../pairtree_root/...' URLs pointing at a given object.

        -  I{"foobar" --> "file:///opt/pairtree_root/fo/ob/ar/obj"}

    @param id: Identifier for a pairtree object
    @type id: identifier
    @param pairtree_root: required path to the pairtree store (and must exist)
    @returns: A 'file:///' URL
    """
    if os.path.isdir(pairtree_root):
        root = os.path.abspath(pairtree_root)
        obj_path = id_to_dirpath(id, shorty_length = shorty_length)
        return "file://%s" % (os.path.join(root, obj_path))
    else:
        raise OSError, "pairtree_root '' does not exist." % pairtree_root


def id_to_dir_list(id, pairtree_root="", shorty_length=2):
    """
    Internal - method for turning an identifier into a list of pairtree 
    directory tree of shorties.

        -  I{"foobar://ark.1" --> ["fo","ob","ar","+=","ar","k,","1"]}

    @param id: Identifer for a pairtree object
    @type id: identifier
    @returns: A list of directory path fragments to the object's root directory
    """
    enc_id = id_encode(id)
    dirpath = []
    if pairtree_root:
        dirpath = [pairtree_root]
    while enc_id:
        dirpath.append(enc_id[:shorty_length])
        enc_id = enc_id[shorty_length:]
    dirpath.append(get_terminator(shorty_length=shorty_length))
    return dirpath




from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup, find_packages

setup(name="Pairtree",
      version="0.7.1-T",
      description="Pairtree FS implementation.",
      long_description="""\
From http://www.cdlib.org/inside/diglib/pairtree/pairtreespec.html : Pairtree, a filesystem hierarchy for holding objects that are located by mapping identifier strings to object directory (or folder) paths two characters at a time. If an object directory (folder) holds all the files, and nothing but the files, that comprise the object, a "pairtree" can be imported by a system that knows nothing about the nature or structure of the objects but can still deliver any object's files by requested identifier. The mapping is reversible, so the importing system can also walk the pairtree and reliably enumerate all the contained object identifiers. To the extent that object dependencies are stored inside the pairtree (e.g., fast indexes stored outside contain only derivative data), simple or complex collections built on top of pairtrees can recover from index failures and reconstruct a collection view simply by walking the trees. Pairtrees have the advantage that many object operations, including backup and restore, can be performed with native operating system tools.
""",
      author="Ben O'Steen",
      author_email="bosteen@gmail.com",
      url="http://packages.python.org/Pairtree/",
      scripts = ['bin/ppath'],
      license="http://www.apache.org/licenses/LICENSE-2.0",
      packages=find_packages(),
      test_suite = "tests.test.TestPairtree",
      )

